from __future__ import annotations  # this is the line this test file is all about

import dataclasses
from enum import Enum
from typing import List, Tuple

import pytest

from serde import SerdeError, deserialize, from_dict, serde, serialize, to_dict
from serde.compat import dataclass_fields


class Status(Enum):
    OK = "ok"
    ERR = "err"


@serde
class A:
    a: int
    b: Status
    c: List[str]


@serde
class B:
    a: A
    b: Tuple[str, A]
    c: Status


# only works with global classes
def test_serde_with_lazy_type_annotations():
    a = A(1, Status.ERR, ["foo"])
    a_dict = {"a": 1, "b": "err", "c": ["foo"]}

    assert a == from_dict(A, a_dict)
    assert a_dict == to_dict(a)

    b = B(a, ("foo", a), Status.OK)
    b_dict = {"a": a_dict, "b": ("foo", a_dict), "c": "ok"}

    assert b == from_dict(B, b_dict)
    assert b_dict == to_dict(b)


# test_forward_reference_works currently only works with global visible classes
@dataclasses.dataclass
class ForwardReferenceFoo:
    # this is not a string forward reference because we use PEP 563 (see 1st line of this file)
    bar: ForwardReferenceBar


@serde
class ForwardReferenceBar:
    i: int


# assert type is str
assert 'ForwardReferenceBar' == dataclasses.fields(ForwardReferenceFoo)[0].type

# setup pyserde for Foo after Bar becomes visible to global scope
deserialize(ForwardReferenceFoo)
serialize(ForwardReferenceFoo)

# now the type really is of type Bar
assert ForwardReferenceBar == dataclasses.fields(ForwardReferenceFoo)[0].type
assert ForwardReferenceBar == next(dataclass_fields(ForwardReferenceFoo)).type


# verify usage works
def test_forward_reference_works():
    h = ForwardReferenceFoo(bar=ForwardReferenceBar(i=10))
    h_dict = {"bar": {"i": 10}}

    assert to_dict(h) == h_dict
    assert from_dict(ForwardReferenceFoo, h_dict) == h


# trying to use forward reference normally will throw
def test_unresolved_forward_reference_throws():
    with pytest.raises(SerdeError) as e:

        @serde
        class UnresolvedForwardFoo:
            bar: UnresolvedForwardBar

        @serde
        class UnresolvedForwardBar:
            i: int

    assert "Failed to resolve type hints for UnresolvedForwardFoo" in str(e)


# trying to use string forward reference will throw
def test_string_forward_reference_throws():
    with pytest.raises(SerdeError) as e:

        @serde
        class UnresolvedStringForwardFoo:
            # string forward references are not compatible with PEP 563 and will throw
            bar: 'UnresolvedStringForwardBar'

        @serde
        class UnresolvedStringForwardBar:
            i: int

    # message is different between <= 3.8 & >= 3.9
    assert "Failed to resolve " in str(e.value)



from typing import Union

from serde import AdjacentTagging, InternalTagging, Untagged, serde
from serde.json import from_json, to_json


@serde
class Bar:
    b: int


@serde
class Baz:
    b: int


def external_tagging():
    @serde
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Baz(10))
    print(f"Into Json: {to_json(f)}")

    s = '{"a": {"Baz": {"b": 10}}}'
    print(f"From Json: {from_json(Foo, s)}")


def internal_tagging():
    @serde(tagging=InternalTagging("type"))
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Baz(10))
    print(f"Into Json: {to_json(f)}")

    s = '{"a": {"type": "Baz", "b": 10}}'
    print(f"From Json: {from_json(Foo, s)}")


def adjacent_tagging():
    @serde(tagging=AdjacentTagging(tag="type", content="content"))
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Baz(10))
    print(f"Into Json: {to_json(f)}")

    s = '{"a": {"type": "Baz", "content": {"b": 10}}}'
    print(f"From Json: {from_json(Foo, s)}")


def untagged():
    @serde(tagging=Untagged)
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Baz(10))
    print(f"Into Json: {to_json(f)}")

    # Untagged cannot correctly deserialize unions with the same fields.
    s = '{"a": {"b": 10}}'
    print(f"From Json: {from_json(Foo, s)}")


def main():
    print('# external tagging (default)')
    external_tagging()
    print('# internal tagging')
    internal_tagging()
    print('# adjacent tagging')
    adjacent_tagging()
    print('# untagged')
    untagged()


if __name__ == '__main__':
    main()



import logging
from dataclasses import dataclass
from ipaddress import IPv4Address
from typing import Dict, List, Optional, Tuple, Union
from uuid import UUID

import pytest

from serde import SerdeError, from_dict
from serde import init as serde_init
from serde import logger, serde, to_dict
from serde.json import from_json, to_json

logging.basicConfig(level=logging.WARNING)
logger.setLevel(logging.DEBUG)

serde_init(True)


@serde
@dataclass(unsafe_hash=True)
class PriUnion:
    """
    Union Primitives.
    """

    v: Union[int, str, float, bool]


@serde
@dataclass(unsafe_hash=True)
class PriOptUnion:
    """
    Union Primitives.
    """

    v: Union[Optional[int], Optional[str], Optional[float], Optional[bool]]


@serde
@dataclass(unsafe_hash=True)
class ContUnion:
    """
    Union Containers.
    """

    v: Union[Dict[str, int], List[int], List[str]]


def test_union():
    v = PriUnion(10)
    s = '{"v": 10}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)

    v = PriUnion(10.0)
    s = '{"v": 10.0}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)

    v = PriUnion('foo')
    s = '{"v": "foo"}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)

    v = PriUnion(True)
    s = '{"v": true}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)


def test_union_optional():
    v = PriOptUnion(10)
    s = '{"v": 10}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion(None)
    s = '{"v": null}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion("foo")
    s = '{"v": "foo"}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion(10.0)
    s = '{"v": 10.0}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion(False)
    s = '{"v": false}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)


def test_union_containers():
    v = ContUnion([1, 2, 3])
    s = '{"v": [1, 2, 3]}'
    assert s == to_json(v)
    assert v == from_json(ContUnion, s)

    v = ContUnion(['1', '2', '3'])
    s = '{"v": ["1", "2", "3"]}'
    assert s == to_json(v)
    assert v == from_json(ContUnion, s)

    v = ContUnion({'a': 1, 'b': 2, 'c': 3})
    s = '{"v": {"a": 1, "b": 2, "c": 3}}'
    assert s == to_json(v)
    # Note: this only works because Dict[str, int] comes first in Union otherwise a List would win
    assert v == from_json(ContUnion, s)


def test_union_with_complex_types():
    @serde
    class A:
        v: Union[int, IPv4Address, UUID]

    a_int = A(1)
    a_int_json = '{"v": 1}'
    assert to_json(a_int) == a_int_json
    assert from_json(A, a_int_json) == a_int
    assert a_int == from_dict(A, to_dict(a_int))

    a_ip = A(IPv4Address("127.0.0.1"))
    a_ip_json = '{"v": "127.0.0.1"}'
    assert to_json(a_ip) == a_ip_json
    assert from_json(A, a_ip_json) == a_ip
    assert a_ip == from_dict(A, to_dict(a_ip))

    a_uid = A(UUID("a317958e-4cbb-4213-9f23-eaff1563c472"))
    a_uid_json = '{"v": "a317958e-4cbb-4213-9f23-eaff1563c472"}'
    assert to_json(a_uid) == a_uid_json
    assert from_json(A, a_uid_json) == a_uid
    assert a_uid == from_dict(A, to_dict(a_uid))


def test_union_with_complex_types_and_reuse_instances():
    @serde(reuse_instances_default=True)
    class A:
        v: Union[int, IPv4Address, UUID]

    a_int = A(1)
    a_int_roundtrip = from_dict(A, to_dict(a_int))
    assert a_int == a_int_roundtrip
    assert a_int.v is a_int_roundtrip.v

    a_ip = A(IPv4Address("127.0.0.1"))
    a_ip_roundtrip = from_dict(A, to_dict(a_ip))
    assert a_ip == a_ip_roundtrip
    assert a_ip.v is a_ip_roundtrip.v

    a_uid = A(UUID("a317958e-4cbb-4213-9f23-eaff1563c472"))
    a_uid_roundtrip = from_dict(A, to_dict(a_uid))
    assert a_uid == a_uid_roundtrip
    assert a_uid.v is a_uid_roundtrip.v


def test_optional_union_with_complex_types():
    @serde
    class A:
        v: Optional[Union[int, IPv4Address, UUID]]

    a = A(123)
    assert a == from_dict(A, to_dict(a, reuse_instances=False), reuse_instances=False)
    assert a == from_dict(A, to_dict(a, reuse_instances=True), reuse_instances=True)

    a_none = A(None)
    assert a_none == from_dict(A, to_dict(a_none, reuse_instances=False), reuse_instances=False)
    assert a_none == from_dict(A, to_dict(a_none, reuse_instances=True), reuse_instances=True)


def test_union_with_complex_types_in_containers():
    @serde
    class A:
        v: Union[List[IPv4Address], List[UUID]]

    a_ips = A([IPv4Address("127.0.0.1"), IPv4Address("10.0.0.1")])
    assert a_ips == from_dict(A, to_dict(a_ips, reuse_instances=False), reuse_instances=False)
    assert a_ips == from_dict(A, to_dict(a_ips, reuse_instances=True), reuse_instances=True)

    a_uids = A([UUID("9c244009-c60d-452b-a378-b8afdc0c2d90"), UUID("5831dc09-20fe-4433-b476-5866b7143364")])
    assert a_uids == from_dict(A, to_dict(a_uids, reuse_instances=False), reuse_instances=False)
    assert a_uids == from_dict(A, to_dict(a_uids, reuse_instances=True), reuse_instances=True)

    a_empty = A([])
    assert a_empty == from_dict(A, to_dict(a_empty, reuse_instances=False), reuse_instances=False)
    assert a_empty == from_dict(A, to_dict(a_empty, reuse_instances=True), reuse_instances=True)


def test_union_exception_if_nothing_matches():
    @serde
    class A:
        v: Union[IPv4Address, UUID]

    with pytest.raises(SerdeError) as ex1:
        from_dict(A, {"v": "not-ip-or-uuid"})
    assert str(ex1.value) == (
        "Can not deserialize 'not-ip-or-uuid' of type str into Union[IPv4Address, UUID].\n"
        "Reasons:\n"
        " Failed to deserialize into IPv4Address: Expected 4 octets in 'not-ip-or-uuid'\n"
        " Failed to deserialize into UUID: badly formed hexadecimal UUID string"
    )

    with pytest.raises(SerdeError) as ex2:
        from_dict(A, {"v": "not-ip-or-uuid"}, reuse_instances=True)
    assert str(ex2.value) == (
        "Can not deserialize 'not-ip-or-uuid' of type str into Union[IPv4Address, UUID].\n"
        "Reasons:\n"
        " Failed to deserialize into IPv4Address: Expected 4 octets in 'not-ip-or-uuid'\n"
        " Failed to deserialize into UUID: badly formed hexadecimal UUID string"
    )

    with pytest.raises(SerdeError) as ex3:
        from_dict(A, {"v": None})
    # omit reason because it is not the same for all python versions & operating systems
    assert str(ex3.value).startswith("Can not deserialize None of type NoneType into Union[IPv4Address, UUID].")

    with pytest.raises(SerdeError) as ex4:
        to_dict(A("not-ip-or-uuid"))
    assert str(ex4.value) == "Can not serialize 'not-ip-or-uuid' of type str for Union[IPv4Address, UUID]"

    with pytest.raises(SerdeError) as ex5:
        to_dict(A("not-ip-or-uuid"), reuse_instances=True)
    assert str(ex5.value) == "Can not serialize 'not-ip-or-uuid' of type str for Union[IPv4Address, UUID]"

    with pytest.raises(SerdeError) as ex6:
        to_dict(A(None), reuse_instances=True)
    assert str(ex6.value) == "Can not serialize None of type NoneType for Union[IPv4Address, UUID]"


def test_union_in_union():
    @serde
    class A:
        v: Union[UUID, Union[int, str]]

    a_uuid = A(UUID("00611ee9-7ca3-41d3-9607-ea7268e264ea"))
    assert a_uuid == from_dict(A, to_dict(a_uuid, reuse_instances=False), reuse_instances=False)
    assert a_uuid == from_dict(A, to_dict(a_uuid, reuse_instances=True), reuse_instances=True)

    a_int = A(1)
    assert a_int == from_dict(A, to_dict(a_int, reuse_instances=False), reuse_instances=False)
    assert a_int == from_dict(A, to_dict(a_int, reuse_instances=True), reuse_instances=True)

    a_str = A("hello")
    assert a_str == from_dict(A, to_dict(a_str, reuse_instances=False), reuse_instances=False)
    assert a_str == from_dict(A, to_dict(a_str, reuse_instances=True), reuse_instances=True)


def test_union_in_other_type():
    @serde
    class A:
        v: Dict[str, Union[UUID, int]]

    a_uuid = A({"key": UUID("00611ee9-7ca3-41d3-9607-ea7268e264ea")})
    assert a_uuid == from_dict(A, to_dict(a_uuid, reuse_instances=False), reuse_instances=False)
    assert a_uuid == from_dict(A, to_dict(a_uuid, reuse_instances=True), reuse_instances=True)

    a_int = A({"key": 1})
    assert a_int == from_dict(A, to_dict(a_int, reuse_instances=False), reuse_instances=False)
    assert a_int == from_dict(A, to_dict(a_int, reuse_instances=True), reuse_instances=True)


def test_union_rename_all():
    @serde(rename_all='pascalcase')
    class Foo:
        bar_baz: Union[int, str]

    assert to_dict(Foo(10)) == {'BarBaz': 10}
    assert from_dict(Foo, {'BarBaz': 'foo'}) == Foo('foo')


def test_union_with_list_of_other_class():
    @serde
    class A:
        a: int

    @serde
    class B:
        b: Union[List[A], str]

    b = B([A(1)])
    b_dict = {"b": [{"a": 1}]}
    assert to_dict(b) == b_dict
    assert from_dict(B, b_dict) == b


# relates to https://github.com/yukinarit/pyserde/issues/113
def test_union_with_union_in_nested_types():
    @serde
    class A:
        v: Union[UUID, List[Union[UUID, int]]]

    a_uuid = A([UUID("00611ee9-7ca3-41d3-9607-ea7268e264ea")])
    assert to_dict(a_uuid, reuse_instances=False) == {"v": ["00611ee9-7ca3-41d3-9607-ea7268e264ea"]}
    assert a_uuid == from_dict(A, to_dict(a_uuid, reuse_instances=False), reuse_instances=False)
    assert a_uuid == from_dict(A, to_dict(a_uuid, reuse_instances=True), reuse_instances=True)

    a_int = A([1])
    assert to_dict(a_int) == {"v": [1]}
    assert a_int == from_dict(A, to_dict(a_int, reuse_instances=False), reuse_instances=False)
    assert a_int == from_dict(A, to_dict(a_int, reuse_instances=True), reuse_instances=True)


# relates to https://github.com/yukinarit/pyserde/issues/113
def test_union_with_union_in_nested_tuple():
    @serde
    class A:
        v: Union[bool, Tuple[Union[str, int]]]

    a_bool = A(False)
    a_bool_dict = {"v": False}
    assert to_dict(a_bool) == a_bool_dict
    assert from_dict(A, a_bool_dict) == a_bool

    a_str = A(("a",))
    a_str_dict = {"v": ("a",)}
    assert to_dict(a_str) == a_str_dict
    assert from_dict(A, a_str_dict) == a_str

    a_int = A((1,))
    a_int_dict = {"v": (1,)}
    assert to_dict(a_int) == a_int_dict
    assert from_dict(A, a_int_dict) == a_int


def test_external_tagging():
    @serde
    class Bar:
        b: int

    @serde
    class Baz:
        b: int

    @serde
    class Nested:
        v: Union[Bar, Baz]

    @serde
    class Foo:
        a: Union[Bar, Baz]
        b: Union[int, str]
        c: Union[Dict[int, str], List[int]]
        d: Union[int, Nested]

    f = Foo(Bar(10), 'foo', {10: 'bar'}, Nested(Baz(100)))
    d = {
        "a": {"Bar": {"b": 10}},  # Union of dataclasses will be (de)serialized with external tagging
        "b": "foo",  # non dataclass will be untagged
        "c": {10: "bar"},
        "d": {"Nested": {"v": {"Baz": {"b": 100}}}},
    }
    assert to_dict(f) == d
    assert from_dict(Foo, d) == f

    @serde
    class Foo:
        a: Union[Bar, int]  # Mix of dataclass and non dataclass

    f = Foo(Bar(10))
    assert from_dict(Foo, to_dict(f)) == f
    f = Foo(10)
    assert from_dict(Foo, to_dict(f)) == f

    @serde
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Bar(10))

    # Tag not found
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"TagNotFound": {"b": 10}}})

    # Tag is correct, but incompatible data
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"Bar": {"c": 10}}})


def test_internal_tagging():
    from serde import InternalTagging

    @serde
    class Bar:
        v: int

    @serde
    class Baz:
        v: int

    @serde(tagging=InternalTagging("type"))
    class Nested:
        v: Union[Bar, Baz]

    @serde(tagging=InternalTagging("type"))
    class Foo:
        a: Union[Bar, Baz]
        b: Union[int, str]
        c: Union[Dict[int, str], List[int]]
        d: Union[int, Nested]

    f = Foo(Bar(10), 'foo', {10: 'bar'}, Nested(Baz(100)))
    d = {
        "a": {"type": "Bar", "v": 10},  # Union of dataclasses will be (de)serialized with internal tagging
        "b": "foo",  # non dataclass will be untagged
        "c": {10: "bar"},
        "d": {"type": "Nested", "v": {"type": "Baz", "v": 100}},
    }
    assert to_dict(f) == d
    assert from_dict(Foo, d) == f

    @serde(tagging=InternalTagging("type"))
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Bar(10))

    # Tag not found
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"TagNotFound": "", "v": 10}})

    # Tag is correct, but incompatible data
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"type": "Bar", "c": 10}})

    with pytest.raises(SerdeError):
        # Tag is not specified in attribute
        @serde(tagging=InternalTagging())
        class Foo:
            pass


def test_adjacent_tagging():
    from serde import AdjacentTagging

    @serde
    class Bar:
        v: int

    @serde
    class Baz:
        v: int

    @serde(tagging=AdjacentTagging("type", "content"))
    class Nested:
        v: Union[Bar, Baz]

    @serde(tagging=AdjacentTagging("type", "content"))
    class Foo:
        a: Union[Bar, Baz]
        b: Union[int, str]
        c: Union[Dict[int, str], List[int]]
        d: Union[int, Nested]

    f = Foo(Bar(10), 'foo', {10: 'bar'}, Nested(Baz(100)))
    d = {
        "a": {"type": "Bar", "content": {"v": 10}},  # Union of dataclasses will be (de)serialized with adjacent tagging
        "b": "foo",  # non dataclass will be untagged
        "c": {10: "bar"},
        "d": {"type": "Nested", "content": {"v": {"type": "Baz", "content": {"v": 100}}}},
    }
    assert to_dict(f) == d
    assert from_dict(Foo, d) == f

    @serde(tagging=AdjacentTagging("type", "content"))
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Bar(10))

    # Tag not found
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"TagNotFound": "", "content": {"v": 10}}})

    # Content tag not found
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"type": "Bar", "TagNotFound": {"v": 10}}})

    # Tag is correct, but incompatible data
    with pytest.raises(Exception):
        assert from_dict(Foo, {"a": {"type": "Bar", "content": {"c": 10}}})

    with pytest.raises(SerdeError):
        # Tag is not specified in attribute
        @serde(tagging=AdjacentTagging(content="content"))
        class Foo:
            pass

    with pytest.raises(SerdeError):
        # Content is not specified in attribute
        @serde(tagging=AdjacentTagging(tag="tag"))
        class Foo:
            pass

    with pytest.raises(SerdeError):
        # Tag/Content is not specified in attribute
        @serde(tagging=AdjacentTagging())
        class Foo:
            pass


def test_untagged():
    from serde import Untagged

    @serde
    class Bar:
        v: int

    @serde
    class Baz:
        v: int

    @serde(tagging=Untagged)
    class Nested:
        v: Union[Bar, Baz]

    @serde(tagging=Untagged)
    class Foo:
        a: Union[Bar, Baz]
        b: Union[int, str]
        c: Union[Dict[int, str], List[int]]
        d: Union[int, Nested]

    f = Foo(Bar(10), 'foo', {10: 'bar'}, Nested(Bar(100)))
    d = {"a": {"v": 10}, "b": "foo", "c": {10: "bar"}, "d": {"v": {"v": 100}}}
    assert to_dict(f) == d
    assert from_dict(Foo, d) == f

    @serde(tagging=Untagged)
    class Foo:
        a: Union[Bar, Baz]

    f = Foo(Baz(10))

    # Untagged can't differenciate the dataclass with similar fields
    with pytest.raises(Exception):
        assert to_dict(from_dict(Foo, d)) == f

    # Untaggled can't differenciate Dict and List.
    @serde
    class Foo:
        a: Union[List[int], Dict[int, str]]

    f = Foo({10: 'bar'})
    with pytest.raises(Exception):
        assert to_dict(from_dict(Foo, d)) == f



from decimal import Decimal
from typing import List, Tuple, Union

from serde.de import from_obj


def test_from_obj():
    # None will be None
    assert from_obj(int, None, False, False) is None

    # Primitive
    assert 10 == from_obj(int, 10, False, False)

    # Union
    assert "a" == from_obj(Union[int, str], "a", False, False)

    # Tuple
    assert ("a", "b") == from_obj(Tuple[str, str], ("a", "b"), False, False)

    # pyserde converts to the specified type
    assert 10 == from_obj(int, "10", False, False)

    # pyserde can converts for container types e.g. Tuple, List etc.
    assert (1, 2) == from_obj(Tuple[int, int], ("1", "2"), False, False)

    # Decimal
    dec = from_obj(List[Decimal], ("0.1", 0.1), False, False)
    assert isinstance(dec[0], Decimal) and dec[0] == Decimal("0.1")
    assert isinstance(dec[1], Decimal) and dec[1] == Decimal(0.1)



"""
Tests for flatten attribute.
"""

from typing import Dict, List

import pytest

from serde import field, serde
from serde.json import from_json, to_json

from .common import all_formats


def test_flatten_simple():
    @serde
    class Bar:
        c: float
        d: bool

    @serde
    class Foo:
        a: int
        b: str
        bar: Bar = field(flatten=True)

    f = Foo(a=10, b='foo', bar=Bar(c=100.0, d=True))
    s = '{"a": 10, "b": "foo", "c": 100.0, "d": true}'
    assert to_json(f) == s
    assert from_json(Foo, s) == f


@pytest.mark.parametrize('se,de', all_formats)
def test_flatten(se, de):
    @serde
    class Baz:
        e: List[int]
        f: Dict[str, str]

    @serde
    class Bar:
        c: float
        d: bool
        baz: Baz = field(flatten=True)

    @serde
    class Foo:
        a: int
        b: str
        bar: Bar = field(flatten=True)

    f = Foo(a=10, b='foo', bar=Bar(c=100.0, d=True, baz=Baz([1, 2], {"a": 10})))
    assert de(Foo, se(f)) == f



from serde import serde, SerdeSkip
from serde.json import from_json
from typing import Optional


def des(cls, o):
    raise SerdeSkip()


@serde(deserializer=des)
class Foo:
    a: Optional[str]
    b: str


if __name__ == "__main__":
    print(from_json(Foo, '{"a": "foo", "b": "bar"}'))  # -> Foo(a='foo')
    print(from_json(Foo, '{}'))           # -> KeyError: 'a'


"""
def from_dict(data, reuse_instances = True):
  if reuse_instances is Ellipsis:
    reuse_instances = True

  if data is None:
    return None

  return cls(
    serde_custom_class_deserializer(
      Optional[str],
      data,
      data.get("a"),
      default=lambda: (serde_custom_class_deserializer(str, data, data.get("a"), default=lambda: data["a"])) if data.get("a") is not None else None),
  )
"""



"""
Serialize and Deserialize in MsgPack format. This module depends on
[msgpack](https://pypi.org/project/msgpack/) package.
"""
from typing import Any, Dict, Type

import msgpack

from .compat import T
from .core import SerdeError
from .de import Deserializer, from_dict, from_tuple
from .se import Serializer, to_dict, to_tuple

__all__ = ["from_msgpack", "to_msgpack"]


class MsgPackSerializer(Serializer):
    @classmethod
    def serialize(cls, obj, use_bin_type: bool = True, ext_type_code: int = None, **opts) -> bytes:
        if ext_type_code is not None:
            obj_bytes = msgpack.packb(obj, use_bin_type=use_bin_type, **opts)
            obj_or_ext = msgpack.ExtType(ext_type_code, obj_bytes)
        else:
            obj_or_ext = obj
        return msgpack.packb(obj_or_ext, use_bin_type=use_bin_type, **opts)


class MsgPackDeserializer(Deserializer):
    @classmethod
    def deserialize(cls, s, raw: bool = False, use_list: bool = False, **opts):
        return msgpack.unpackb(s, raw=raw, use_list=use_list, **opts)


def to_msgpack(
    obj: Any, se: Type[Serializer] = MsgPackSerializer, named: bool = True, ext_dict: Dict[Type, int] = None, **opts
) -> bytes:
    """
    Serialize the object into MsgPack.

    You can pass any serializable `obj`. If `ext_dict` option is specified, `obj` is encoded as a `msgpack.ExtType`
    If you supply other keyword arguments, they will be passed in `msgpack.packb` function.

    If `named` is True, field names are preserved, namely the object is encoded as `dict` then serialized into MsgPack.
    If `named` is False, the object is encoded as `tuple` then serialized into MsgPack. `named=False` will produces
    compact binary.

    If you want to use the other msgpack package, you can subclass `MsgPackSerializer` and implement your own logic.
    """
    ext_type_code = None
    if ext_dict is not None:
        obj_type = type(obj)
        ext_type_code = ext_dict.get(obj_type)
        if ext_type_code is None:
            raise SerdeError(f"Could not find type code for {obj_type.__name__} in ext_dict")

    to_func = to_dict if named else to_tuple
    return se.serialize(to_func(obj, reuse_instances=False, convert_sets=True), ext_type_code=ext_type_code, **opts)


def from_msgpack(
    c: Type[T],
    s: str,
    de: Type[Deserializer] = MsgPackDeserializer,
    named: bool = True,
    ext_dict: Dict[int, Type] = None,
    **opts,
) -> Type[T]:
    """
    Deserialize from MsgPack into the object.

    `c` is a class obejct and `s` is MsgPack binary. If `ext_dict` option is specified, `c` is ignored and type is
    inferred from `msgpack.ExtType` If you supply other keyword arguments, they will be passed in
    `msgpack.unpackb` function.

    If you want to use the other msgpack package, you can subclass `MsgPackDeserializer` and implement your own logic.
    """
    if ext_dict is not None:
        ext = de.deserialize(s, **opts)
        ext_type = ext_dict.get(ext.code)
        if ext_type is None:
            raise SerdeError(f"Could not find type for code {ext.code} in ext_dict")
        return from_msgpack(ext_type, ext.data, de, named, **opts)
    else:
        from_func = from_dict if named else from_tuple
        return from_func(c, de.deserialize(s, **opts), reuse_instances=False)



"""
pyserde inspection tool.

#### Usage

```
$ python -m serde.inspect <PATH> <NAME>

 PATH  Python script path.
 NAME  Pyserde class name.
```

"""
import argparse
import importlib
import logging
import os
import sys
from typing import Type

from .core import SERDE_SCOPE, SerdeScope, init, logger

init(True)


def inspect(cls: Type) -> None:
    """
    Inspect a pyserde class.
    """
    scope: SerdeScope = getattr(cls, SERDE_SCOPE, {})
    print(scope)


def main(arg):
    """
    Main entrypoint of `serde.inspect`.
    """
    if arg.verbose:
        logging.basicConfig(level=logging.DEBUG)

    try:
        import black

        assert black
    except ImportError:
        logger.warning(('Tips: Installing "black" makes the output prettier! Try this command:\n' 'pip install back'))

    dir = os.path.dirname(arg.path)
    mod = os.path.basename(arg.path)[:-3]
    print(f'Loading {mod}.{arg.name} from {dir}/{mod}.py')
    sys.path.append(dir)
    pkg = importlib.import_module(mod)
    cls = getattr(pkg, arg.name)
    inspect(cls)


parser = argparse.ArgumentParser(description='pyserde-inspect')
parser.add_argument('path', type=str, help='Python script path.')
parser.add_argument('name', type=str, help='Pyserde class name.')
parser.add_argument('-v', dest='verbose', action='store_true', help='Enable debug logging.')


if __name__ == '__main__':
    args = parser.parse_args()
    main(args)



"""
tomlfile.py

Read Pipenv Pipfile by pyserde.

Usage:
    $ poetry install
    $ poetry run python tomlfile.py
"""
from typing import Dict, List, Optional, Union

from serde import Untagged, serde
from serde.toml import from_toml


@serde
class Source:
    url: str
    verify_ssl: bool
    name: str


@serde
class Requires:
    python_version: str


@serde
class Package:
    path: Optional[str] = None
    version: Optional[str] = None
    editable: Optional[bool] = False


@serde(tagging=Untagged)
class Pipfile:
    source: List[Source]
    requires: Optional[Requires]
    packages: Dict[str, Union[str, Package]]


def main():
    with open('Pipfile') as f:
        toml = f.read()
    pip = from_toml(Pipfile, toml)
    print(pip)


if __name__ == '__main__':
    main()



from flask import Flask, Response, request

from serde import serde
from serde.json import from_json, to_json


@serde
class ToDo:
    id: int
    title: str
    description: str
    done: bool


app = Flask(__name__)


@app.route('/')
def hello_world():
    return 'Hello, World!'


@app.route('/todos', methods=['GET', 'POST'])
def todos():
    print(request.method)
    if request.method == 'GET':
        body = to_json([ToDo(1, 'Play games', 'Play 聖剣伝説3', False)])
        return Response(body, mimetype='application/json')
    else:
        todo = from_json(ToDo, request.get_data())
        return f'A new ToDo {todo} successfully created.'


if __name__ == '__main__':
    app.run(debug=True)



from datetime import datetime

from serde import SerdeSkip, default_deserializer, default_serializer, field, serde
from serde.json import from_json, to_json


def serializer(cls, o):
    """
    Custom class level serializer.
    """

    # You can provide a custom serialize/deserialize logic for certain types.
    if cls is datetime:
        return o.strftime('%d/%m/%y')
    # Raise SerdeSkip to tell serde to use the default serializer/deserializer.
    else:
        raise SerdeSkip()


def deserializer(cls, o):
    """
    Custom class level deserializer.
    """
    if cls is datetime:
        return datetime.strptime(o, '%d/%m/%y')
    else:
        raise SerdeSkip()


@serde(serializer=serializer, deserializer=deserializer)
class Foo:
    i: int
    dt1: datetime
    # Override by field serializer/deserializer.
    dt2: datetime = field(
        serializer=lambda x: x.strftime('%y.%m.%d'), deserializer=lambda x: datetime.strptime(x, '%y.%m.%d')
    )
    # Override by the default serializer/deserializer.
    dt3: datetime = field(serializer=default_serializer, deserializer=default_deserializer)


def main():
    dt = datetime(2021, 1, 1, 0, 0, 0)
    f = Foo(10, dt, dt, dt)
    print(f"Into Json: {to_json(f)}")

    s = '{"i": 10, "dt1": "01/01/21", "dt2": "01.01.21", "dt3": "2021-01-01T00:00:00"}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



"""
rename_all.py


Usage:
    $ poetry install
    $ poetry run python rename_all.py
"""
from typing import Optional

from serde import serde
from serde.json import from_json, to_json


@serde(rename_all='pascalcase')
class Foo:
    name: str
    no: Optional[int] = None


def main():
    f = Foo('Pikachu')
    print(f"Into Json: {to_json(f)}")

    s = '{"Name": "Pikachu", "No": 25}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



"""
Serialize and Deserialize in JSON format.
"""
import json
from typing import Any, Type

from .compat import T
from .de import Deserializer, from_dict
from .se import Serializer, to_dict

__all__ = ["from_json", "to_json"]


class JsonSerializer(Serializer):
    @classmethod
    def serialize(cls, obj: Any, **opts) -> str:
        return json.dumps(obj, **opts)


class JsonDeserializer(Deserializer):
    @classmethod
    def deserialize(cls, s, **opts):
        return json.loads(s, **opts)


def to_json(obj: Any, se: Type[Serializer] = JsonSerializer, **opts) -> str:
    """
    Serialize the object into JSON.

    You can pass any serializable `obj`. If you supply other keyword arguments, they will be passed in `json.dumps`
    function.

    If you want to use the other json package, you can subclass `JsonSerializer` and implement your own logic.
    """
    return se.serialize(to_dict(obj, reuse_instances=False, convert_sets=True), **opts)


def from_json(c: Type[T], s: str, de: Type[Deserializer] = JsonDeserializer, **opts) -> T:
    """
    Deserialize from JSON into the object.

    `c` is a class obejct and `s` is JSON string.  If you supply other keyword arguments, they will be passed in
    `json.loads` function.

    If you want to use the other json package, you can subclass `JsonDeserializer` and implement your own logic.
    """
    return from_dict(c, de.deserialize(s, **opts), reuse_instances=False)



from typing import Dict

from serde import field, serde
from serde.json import from_json, to_json


@serde
class Foo:
    i: int = 10
    s: str = 'foo'
    f: float = field(default=100.0)  # Use dataclass field.
    b: bool = field(default=True)
    d: Dict[str, int] = field(default_factory=dict)


def main():
    h = Foo()
    print(f"Into Json: {to_json(h)}")

    s = '{}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



import sys

import any
import collection
import custom_class_serializer
import custom_field_serializer
import default
import env
import flatten
import forward_reference
import jsonfile
import newtype
import rename
import rename_all
import simple
import skip
import tomlfile
import type_datetime
import type_decimal
import union
import union_tagging
import yamlfile

PY36 = sys.version_info[:3] < (3, 7, 0)

if not PY36:
    import lazy_type_evaluation


def run_all():
    run(any)
    run(simple)
    run(newtype)
    run(collection)
    run(default)
    run(env)
    run(flatten)
    run(jsonfile)
    run(rename)
    run(rename_all)
    run(skip)
    run(tomlfile)
    run(yamlfile)
    run(union)
    run(custom_class_serializer)
    run(custom_field_serializer)
    run(forward_reference)
    run(type_decimal)
    run(type_datetime)
    if not PY36:
        run(lazy_type_evaluation)


def run(module):
    print('-----------------')
    print(f'running {module.__name__}')
    module.main()


if __name__ == '__main__':
    try:
        run_all()
    except Exception:
        sys.exit(1)



import datetime
import decimal
import ipaddress
import itertools
import os
import pathlib
import sys
import uuid
from typing import Any, Dict, List, NewType, Optional, Set, Tuple

import more_itertools

from serde import from_dict, from_tuple, to_dict, to_tuple
from serde.json import from_json, to_json
from serde.msgpack import from_msgpack, to_msgpack
from serde.toml import from_toml, to_toml
from serde.yaml import from_yaml, to_yaml
from tests import data

format_dict: List = [(to_dict, from_dict)]

format_tuple: List = [(to_tuple, from_tuple)]

format_json: List = [(to_json, from_json)]

format_msgpack: List = [(to_msgpack, from_msgpack)]

format_yaml: List = [(to_yaml, from_yaml)]

format_toml: List = [(to_toml, from_toml)]

all_formats: List = format_dict + format_tuple + format_json + format_msgpack + format_yaml + format_toml

types: List = [
    (10, int),  # Primitive
    ('foo', str),
    (100.0, float),
    (True, bool),
    (10, Optional[int]),  # Optional
    (None, Optional[int]),
    ([1, 2], List[int]),  # Container
    ([1, 2], List),
    ([1, 2], list),
    ([], List[int]),
    ({1, 2}, Set[int]),
    ({1, 2}, Set),
    ({1, 2}, set),
    (set(), Set[int]),
    ((1, 1), Tuple[int, int]),
    ((1, 1), Tuple),
    ({'a': 1}, Dict[str, int]),
    ({'a': 1}, Dict),
    ({'a': 1}, dict),
    ({}, Dict[str, int]),
    (data.Pri(10, 'foo', 100.0, True), data.Pri),  # dataclass
    (data.Pri(10, 'foo', 100.0, True), Optional[data.Pri]),
    (None, Optional[data.Pri]),
    (10, NewType('Int', int)),  # NewType
    ({'a': 1}, Any),  # Any
    (pathlib.Path('/tmp/foo'), pathlib.Path),  # Extended types
    (pathlib.Path('/tmp/foo'), Optional[pathlib.Path]),
    (None, Optional[pathlib.Path]),
    (pathlib.PurePath('/tmp/foo'), pathlib.PurePath),
    (pathlib.PurePosixPath('/tmp/foo'), pathlib.PurePosixPath),
    (pathlib.PureWindowsPath('C:\\tmp'), pathlib.PureWindowsPath),
    (uuid.UUID("8f85b32c-a0be-466c-87eb-b7bbf7a01683"), uuid.UUID),
    (ipaddress.IPv4Address("127.0.0.1"), ipaddress.IPv4Address),
    (ipaddress.IPv6Address("::1"), ipaddress.IPv6Address),
    (ipaddress.IPv4Network("127.0.0.0/8"), ipaddress.IPv4Network),
    (ipaddress.IPv6Network("::/128"), ipaddress.IPv6Network),
    (ipaddress.IPv4Interface("192.168.1.1/24"), ipaddress.IPv4Interface),
    (ipaddress.IPv6Interface("::1/128"), ipaddress.IPv6Interface),
    (decimal.Decimal(10), decimal.Decimal),
    (datetime.datetime.strptime('Jan 1 2021 1:55PM', '%b %d %Y %I:%M%p'), datetime.datetime),
    (datetime.datetime.strptime('Jan 1 2021 1:55PM', '%b %d %Y %I:%M%p').date(), datetime.date),
    (datetime.datetime.strptime('Jan 1 2021 1:55PM', '%b %d %Y %I:%M%p').time(), datetime.time),
]

# these types can only be instantiated on their corresponding system
if os.name == "posix":
    types.append((pathlib.PosixPath('/tmp/foo'), pathlib.PosixPath))
if os.name == "nt":
    types.append((pathlib.WindowsPath('C:\\tmp'), pathlib.WindowsPath))

if sys.version_info[:3] >= (3, 9, 0):
    types.extend([([1, 2], list[int]), ({'a': 1}, dict[str, int]), ((1, 1), tuple[int, int])])

types_combinations: List = list(map(lambda c: list(more_itertools.flatten(c)), itertools.combinations(types, 2)))

opt_case: List = [
    {'reuse_instances_default': False},
    {'reuse_instances_default': False, 'rename_all': 'camelcase'},
    {'reuse_instances_default': False, 'rename_all': 'snakecase'},
]


def make_id_from_dict(d: Dict) -> str:
    if not d:
        return 'none'
    else:
        key = list(d)[0]
        return f'{key}-{d[key]}'


def opt_case_ids():
    return list(map(make_id_from_dict, opt_case))


def type_ids():
    from serde.compat import typename

    def make_id(pair: Tuple):
        t, T = pair
        return f'{typename(T)}({t})'

    return list(map(make_id, types))


def type_combinations_ids():
    from serde.compat import typename

    def make_id(quad: Tuple):
        t, T, u, U = quad
        return f'{typename(T)}({t})-{typename(U)}({u})'

    return list(map(make_id, types_combinations))



import timeit
from dataclasses import dataclass


@dataclass
class Foo:
    a: int
    b: float
    c: str
    d: bool

    def to_dict(self):
        return {"a": self.a, "b": self.b, "c": self.c, "d": self.d}

    def to_dict_typecheck(self):
        if not isinstance(self.a, int):
            raise Exception("'a' is not type of int")
        if not isinstance(self.b, float):
            raise Exception("'b' is not type of float")
        if not isinstance(self.c, str):
            raise Exception("'c' is not type of str")
        if not isinstance(self.d, bool):
            raise Exception("'d' is not type of bool")
        return {"a": self.a, "b": self.b, "c": self.c, "d": self.d}

    def to_dict_conversion(self):
        return {"a": int(self.a), "b": float(self.b), "c": str(self.c), "d": bool(self.d)}

    @classmethod
    def from_dict(cls, data):
        return cls(a=data["a"], b=data["b"], c=data["c"], d=["d"])

    @classmethod
    def from_dict_typecheck(cls, data):
        if not isinstance(data["a"], int):
            raise Exception("'a' is not type of int")
        if not isinstance(data["b"], float):
            raise Exception("'b' is not type of float")
        if not isinstance(data["c"], str):
            raise Exception("'c' is not type of str")
        if not isinstance(data["d"], bool):
            raise Exception("'d' is not type of bool")
        return cls(a=data["a"], b=data["b"], c=data["c"], d=["d"])

    @classmethod
    def from_dict_conversion(cls, data):
        return cls(a=int(data["a"]), b=float(data["b"]), c=str(data["c"]), d=bool(["d"]))


def main():
    f = Foo(10, 10.0, "foo", True)
    data = {"a": 10, "b": 10.0, "c": "foo", "d": True}
    print(timeit.timeit(lambda: f.to_dict(), number=10000000))
    print(timeit.timeit(lambda: f.to_dict_typecheck(), number=10000000))
    print(timeit.timeit(lambda: f.to_dict_conversion(), number=10000000))
    print(timeit.timeit(lambda: Foo.from_dict(data), number=10000000))
    print(timeit.timeit(lambda: Foo.from_dict_typecheck(data), number=10000000))
    print(timeit.timeit(lambda: Foo.from_dict_conversion(data), number=10000000))


if __name__ == "__main__":
    main()



from typing import Dict, List, Union

from serde import serde
from serde.json import from_json, to_json


@serde
class Bar:
    v: int


@serde
class Baz:
    v: float


@serde
class Foo:
    a: Union[int, str]
    b: Union[Dict[str, int], List[int]]
    c: Union[Bar, Baz]


def main():
    f = Foo(10, [1, 2, 3], Bar(10))
    print(f"Into Json: {to_json(f)}")

    s = '{"a": 10, "b": [1, 2, 3], "c": {"Bar": {"v": 10}}}'
    print(f"From Json: {from_json(Foo, s)}")

    f = Foo('foo', {'bar': 1, 'baz': 2}, Baz(100.0))
    print(f"Into Json: {to_json(f)}")

    s = '{"a": "foo", "b": {"bar": 1, "baz": 2}, "c": {"Baz": {"v": 100.0}}}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



"""
rename.py


Usage:
    $ poetry install
    $ poetry run python rename.py
"""

from serde import field, serde
from serde.json import to_json


@serde
class Foo:
    # Use 'class_name' because 'class' is a keyword.
    class_name: str = field(rename='class')


def main():
    print(to_json(Foo(class_name='Foo')))


if __name__ == '__main__':
    main()



import datetime

from serde import serde
from serde.json import from_json, to_json


@serde
class Foo:
    d: datetime.date
    t: datetime.time
    dt: datetime.datetime


def main():
    dt = datetime.datetime(2021, 1, 1, 0, 0, 0)

    foo = Foo(dt.date(), dt.time(), dt)
    print(f"Into Json: {to_json(foo)}")

    s = '{"d": "2021-01-01", "t": "00:00:00", "dt": "2021-01-01T00:00:00"}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



"""
Tests for custom serializer/deserializer.
"""
from datetime import datetime
from typing import Optional, Union

import pytest

from serde import SerdeError, SerdeSkip, default_deserializer, default_serializer, field, from_tuple, serde, to_tuple
from serde.json import from_json, to_json


def test_custom_field_serializer():
    @serde
    class Foo:
        a: datetime
        b: datetime = field(
            serializer=lambda x: x.strftime('%d/%m/%y'), deserializer=lambda x: datetime.strptime(x, '%d/%m/%y')
        )
        c: Optional[datetime] = field(
            serializer=lambda x: x.strftime('%d/%m/%y') if x else None,
            deserializer=lambda x: datetime.strptime(x, '%d/%m/%y') if x else None,
        )

    dt = datetime(2021, 1, 1, 0, 0, 0)
    f = Foo(dt, dt, None)

    assert to_json(f) == '{"a": "2021-01-01T00:00:00", "b": "01/01/21", "c": null}'
    assert f == from_json(Foo, to_json(f))

    assert to_tuple(f) == (datetime(2021, 1, 1, 0, 0), '01/01/21', None)
    assert f == from_tuple(Foo, to_tuple(f))


def test_raise_error():
    def raise_exception(_):
        raise Exception()

    @serde
    class Foo:
        i: int = field(serializer=raise_exception, deserializer=raise_exception)

    f = Foo(10)
    with pytest.raises(Exception):
        to_json(f)

    with pytest.raises(Exception):
        from_json(Foo, '{"i": 10}')


def test_wrong_signature():
    @serde
    class Foo:
        i: int = field(serializer=lambda: '10', deserializer=lambda: 10)

    f = Foo(10)
    with pytest.raises(SerdeError):
        to_json(f)

    with pytest.raises(SerdeError):
        from_json(Foo, '{"i": 10}')


def test_custom_class_serializer():
    def serializer(cls, o):
        if cls is datetime:
            return o.strftime('%d/%m/%y')
        else:
            raise SerdeSkip()

    def deserializer(cls, o):
        if cls is datetime:
            return datetime.strptime(o, '%d/%m/%y')
        else:
            raise SerdeSkip()

    @serde(serializer=serializer, deserializer=deserializer)
    class Foo:
        a: int
        b: datetime
        c: datetime
        d: Optional[str] = None
        e: Union[str, int] = 10

    dt = datetime(2021, 1, 1, 0, 0, 0)
    f = Foo(10, dt, dt)

    assert to_json(f) == '{"a": 10, "b": "01/01/21", "c": "01/01/21", "d": null, "e": 10}'
    assert f == from_json(Foo, to_json(f))

    assert to_tuple(f) == (10, '01/01/21', '01/01/21', None, 10)
    assert f == from_tuple(Foo, to_tuple(f))

    def fallback(_, __):
        raise SerdeSkip()

    @serde(serializer=fallback, deserializer=fallback)
    class Foo:
        a: Optional[str]
        b: str

    f = Foo("foo", "bar")
    assert to_json(f) == '{"a": "foo", "b": "bar"}'
    assert f == from_json(Foo, '{"a": "foo", "b": "bar"}')
    assert Foo(None, "bar") == from_json(Foo, '{"b": "bar"}')
    with pytest.raises(Exception):
        assert Foo(None, "bar") == from_json(Foo, '{}')
    with pytest.raises(Exception):
        assert Foo("foo", "bar") == from_json(Foo, '{"a": "foo"}')


def test_field_serialize_override_class_serializer():
    def serializer(cls, o):
        if cls is datetime:
            return o.strftime('%d/%m/%y')
        else:
            raise SerdeSkip()

    def deserializer(cls, o):
        if cls is datetime:
            return datetime.strptime(o, '%d/%m/%y')
        else:
            raise SerdeSkip()

    @serde(serializer=serializer, deserializer=deserializer)
    class Foo:
        a: int
        b: datetime
        c: datetime = field(
            serializer=lambda x: x.strftime('%y.%m.%d'), deserializer=lambda x: datetime.strptime(x, '%y.%m.%d')
        )

    dt = datetime(2021, 1, 1, 0, 0, 0)
    f = Foo(10, dt, dt)

    assert to_json(f) == '{"a": 10, "b": "01/01/21", "c": "21.01.01"}'
    assert f == from_json(Foo, to_json(f))

    assert to_tuple(f) == (10, '01/01/21', '21.01.01')
    assert f == from_tuple(Foo, to_tuple(f))


def test_override_by_default_serializer():
    def serializer(cls, o):
        if cls is datetime:
            return o.strftime('%d/%m/%y')
        else:
            raise SerdeSkip()

    def deserializer(cls, o):
        if cls is datetime:
            return datetime.strptime(o, '%d/%m/%y')
        else:
            raise SerdeSkip()

    @serde(serializer=serializer, deserializer=deserializer)
    class Foo:
        a: int
        b: datetime
        c: datetime = field(serializer=default_serializer, deserializer=default_deserializer)

    dt = datetime(2021, 1, 1, 0, 0, 0)
    f = Foo(10, dt, dt)

    assert to_json(f) == '{"a": 10, "b": "01/01/21", "c": "2021-01-01T00:00:00"}'
    assert f == from_json(Foo, to_json(f))

    assert to_tuple(f) == (10, '01/01/21', datetime(2021, 1, 1, 0, 0))
    assert f == from_tuple(Foo, to_tuple(f))



from datetime import datetime

from serde import field, serde
from serde.json import from_json, to_json


@serde
class Foo:
    dt1: datetime
    dt2: datetime = field(
        serializer=lambda x: x.strftime('%d/%m/%y'), deserializer=lambda x: datetime.strptime(x, '%d/%m/%y')
    )


def main():
    dt = datetime(2021, 1, 1, 0, 0, 0)
    f = Foo(dt, dt)
    print(f"Into Json: {to_json(f)}")

    s = '{"dt1": "2021-01-01T00:00:00", "dt2": "01/01/21"}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



from serde import field, serde
from serde.json import from_json, to_json


@serde
class Bar:
    c: float
    d: bool


@serde
class Foo:
    a: int
    b: str
    bar: Bar = field(flatten=True)


def main():
    f = Foo(a=10, b='foo', bar=Bar(c=100.0, d=True))
    print(f"Into Json: {to_json(f)}")

    s = '{"a": 10, "b": "foo", "c": 100.0, "d": true}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



"""
jsonfile.py

Make an http request to JSON WebAPI.

Usage:
    $ poetry install
    $ poetry run python jsonfile.py
"""
from typing import List, Optional

import requests

from serde import serde
from serde.json import from_json


@serde
class Slide:
    title: str
    type: str
    items: Optional[List[str]]


@serde
class Slideshow:
    author: str
    date: str
    slides: List[Slide]
    title: str


@serde
class Data:
    slideshow: Slideshow


def main():
    text = requests.get('https://httpbin.org/json').text
    data = from_json(Data, text)
    print(data)


if __name__ == '__main__':
    main()



from dataclasses import dataclass

from envclasses import envclass, load_env

from serde import deserialize
from serde.yaml import from_yaml


@deserialize
@envclass
@dataclass
class App:
    addr: str
    port: int
    secret: str
    workers: int


def main():
    with open('app.yml') as f:
        yml = f.read()
    cfg = from_yaml(App, yml)
    print(cfg)

    load_env(cfg, prefix='APP')
    print(cfg)


if __name__ == '__main__':
    main()



from enum import Enum
from typing import List
from dataclasses import dataclass
from mashumaro import DataClassJSONMixin

class Currency(Enum):
    USD = "USD"
    EUR = "EUR"

@dataclass
class CurrencyPosition(DataClassJSONMixin):
    currency: Currency
    balance: float

@dataclass
class StockPosition(DataClassJSONMixin):
    ticker: str
    name: str
    balance: int

@dataclass
class Portfolio(DataClassJSONMixin):
    currencies: List[CurrencyPosition]
    stocks: List[StockPosition]

my_portfolio = Portfolio(
    currencies=[
        CurrencyPosition(Currency.USD, "238.67"),
        CurrencyPosition(Currency.EUR, "10"),
        StockPosition("AAPL", "Apple", "10"),
    ],
    stocks=[
        StockPosition("AAPL", "Apple", "10"),
        StockPosition("AMZN", "Amazon", "10"),
    ]
)

json_string = my_portfolio.to_json()
print(json_string)
print(Portfolio.from_json(json_string))  # same as my_portfolio



from typing import NewType

from serde import serde
from serde.json import from_json, to_json

UserId = NewType("UserId", int)


@serde
class Foo:
    id: UserId


def main():
    f = Foo(id=UserId(10))
    print(f"Into Json: {to_json(f)}")

    s = '{"id": 10}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



from serde import serde, field
from serde.json import to_json, from_json
from macaddress import EUI48
from typing import Optional


@:serde
class Foo:
    addr: EUI48 = field(serializer=lambda x: str(x), deserializer=lambda x: EUI48(x))
    opt_addr: Optional[EUI48] = field(serializer=lambda x: x and str(x), deserializer=lambda x: x and EUI48(x))


f = Foo(EUI48('01-23-45-67-89-ab'), None)
print(f)
json_data = to_json(f)
print(json_data)
f = from_json(Foo, json_data)
print(f)



"""
`pyserde` is a yet another serialization library on top of
[dataclasses](https://docs.python.org/3/library/dataclasses.html).

## Overview

Declare a class with pyserde's `@serialize` and `@deserialize` decorators.

>>> from serde import serde
>>>
>>> @serde
... class Foo:
...     i: int
...     s: str
...     f: float
...     b: bool

You can serialize `Foo` object into JSON.

>>> from serde.json import to_json
>>> to_json(Foo(i=10, s='foo', f=100.0, b=True))
'{"i": 10, "s": "foo", "f": 100.0, "b": true}'

You can deserialize JSON into `Foo` object.
>>> from serde.json import from_json
>>> from_json(Foo, '{"i": 10, "s": "foo", "f": 100.0, "b": true}')
Foo(i=10, s='foo', f=100.0, b=True)

## Modules

The following modules provide the core functionalities of `pyserde`.
* `serde.se`: All about serialization.
* `serde.de`: All about deserialization.
* `serde.core`: Core module used by `serde.se` and `serde.de`.
* `serde.compat`: Provides compatibility layer which handles typing.

The following modules provides functionalities for supported data formats.
* `serde.json`: Serialize and Deserialize in JSON.
* `serde.msgpack`: Serialize and Deserialize in MsgPack.
* `serde.yaml`: Serialize and Deserialize in YAML.
* `serde.toml`: Serialize and Deserialize in TOML.
"""

import dataclasses
import sys

from .compat import SerdeError, SerdeSkip  # noqa
from .core import AdjacentTagging, ExternalTagging, InternalTagging, Untagged, field, init, logger  # noqa
from .de import default_deserializer, deserialize, from_dict, from_tuple, is_deserializable  # noqa
from .se import asdict, astuple, default_serializer, is_serializable, serialize, to_dict, to_tuple  # noqa

if sys.version_info[:2] == (3, 6):
    import backports.datetime_fromisoformat

    backports.datetime_fromisoformat.MonkeyPatch.patch_fromisoformat()


def serde(_cls=None, **kwargs):
    def wrap(cls):
        dataclasses.dataclass(cls)
        serialize(cls, **kwargs)
        deserialize(cls, **kwargs)
        return cls

    if _cls is None:
        return wrap  # type: ignore

    return wrap(_cls)



"""
skip.py

Example usage of skip and skip_if attributes.

Usage:
    $ poetry install
    $ poetry run python skip.py
"""

from typing import Dict, List

from serde import field, serde
from serde.json import to_json


@serde
class Resource:
    name: str
    hash: str
    metadata: Dict[str, str] = field(default_factory=dict, skip=True)


@serde
class World:
    player: str
    enemies: List[str] = field(default_factory=list, skip_if_false=True)
    buddy: str = field(default='', skip_if=lambda v: v == 'Pikachu')
    town: str = field(default='Masara Town', skip_if_default=True)


def main():
    resources = [
        Resource("Stack Overflow", "b6469c3f31653d281bbbfa6f94d60fea130abe38"),
        Resource("GitHub", "5cb7a0c47e53854cd00e1a968de5abce1c124601", metadata={"headquarters": "San Francisco"}),
    ]
    print(to_json(resources))

    # "buddy" and "town" field will be omitted
    world = World('satoshi', ['Rattata', 'Pidgey'], 'Pikachu')
    print(to_json(world))

    # "enemies" field will be omitted
    world = World('green', [], 'Charmander', 'Black City')
    print(to_json(world))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['examples', 'examples.webapi', 'serde', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['jinja2', 'stringcase', 'typing_inspect>=0.4.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses',
                             'backports-datetime-fromisoformat>=1,<2'],
 'all': ['msgpack', 'toml', 'pyyaml'],
 'msgpack': ['msgpack'],
 'toml': ['toml'],
 'yaml': ['pyyaml']}

setup_kwargs = {
    'name': 'pyserde',
    'version': '0.7.0',
    'description': 'Yet another serialization library on top of dataclasses',
    'long_description': '# `pyserde`\n\nYet another serialization library on top of [dataclasses](https://docs.python.org/3/library/dataclasses.html), inspired by [serde-rs](https://github.com/serde-rs/serde).\n\n[![image](https://img.shields.io/pypi/v/pyserde.svg)](https://pypi.org/project/pyserde/)\n[![image](https://img.shields.io/pypi/pyversions/pyserde.svg)](https://pypi.org/project/pyserde/)\n![Tests](https://github.com/yukinarit/pyserde/workflows/Tests/badge.svg)\n[![codecov](https://codecov.io/gh/yukinarit/pyserde/branch/master/graph/badge.svg)](https://codecov.io/gh/yukinarit/pyserde)\n\n[Guide](https://yukinarit.github.io/pyserde/guide) | [API Docs](https://yukinarit.github.io/pyserde/api/serde.html) | [Examples](./examples)\n\n## Overview\n\nDeclare a class with pyserde\'s `@serde` decorator.\n\n```python\n@serde\nclass Foo:\n    i: int\n    s: str\n    f: float\n    b: bool\n```\n\nYou can serialize `Foo` object into JSON.\n\n```python\n>>> to_json(Foo(i=10, s=\'foo\', f=100.0, b=True))\n{"i": 10, "s": "foo", "f": 100.0, "b": true}\n```\n\nYou can deserialize JSON into `Foo` object.\n```python\n>>> from_json(Foo, \'{"i": 10, "s": "foo", "f": 100.0, "b": true}\')\nFoo(i=10, s=\'foo\', f=100.0, b=True)\n```\n\n## Features\n\n- Supported data formats\n    - dict\n    - tuple\n    - JSON\n\t- Yaml\n\t- Toml\n\t- MsgPack\n- Supported types\n    - Primitives (`int`, `float`, `str`, `bool`)\n    - Containers (`List`, `Set`, `Tuple`, `Dict`)\n    - [`typing.Optional`](https://docs.python.org/3/library/typing.html#typing.Optional)\n    - [`typing.Union`](https://docs.python.org/3/library/typing.html#typing.Union)\n    - User defined class with [`@dataclass`](https://docs.python.org/3/library/dataclasses.html)\n    - [`typing.NewType`](https://docs.python.org/3/library/typing.html#newtype) for primitive types\n    - [`typing.Any`](https://docs.python.org/3/library/typing.html#the-any-type)\n    - [`Enum`](https://docs.python.org/3/library/enum.html#enum.Enum) and [`IntEnum`](https://docs.python.org/3/library/enum.html#enum.IntEnum)\n    - Standard library\n        - [`pathlib.Path`](https://docs.python.org/3/library/pathlib.html)\n        - [`decimal.Decimal`](https://docs.python.org/3/library/decimal.html)\n        - [`uuid.UUID`](https://docs.python.org/3/library/uuid.html)\n        - [`datetime.date`](https://docs.python.org/3/library/datetime.html#date-objects), [`datetime.time`](https://docs.python.org/3/library/datetime.html#time-objects), [`datetime.datetime`](https://docs.python.org/3/library/datetime.html#datetime-objects)\n        - [`ipaddress`](https://docs.python.org/3/library/ipaddress.html)\n- [Attributes](docs/features/attributes.md)\n- [Decorators](docs/features/decorators.md)\n- [Union Representation](docs/features/union.md)\n- [Python 3.9 type hinting](docs/features/python3.9-type-hinting.md)\n- [Postponed evaluation of type annotation](docs/features/postponed-evaluation-of-type-annotation.md)\n- [Forward reference](docs/features/forward-reference.md)\n- [Case Conversion](docs/features/case-conversion.md)\n- [Rename](docs/features/rename.md)\n- [Skip](docs/features/skip.md)\n- [Conditional Skip](docs/features/conditional-skip.md)\n- [Custom field (de)serializer](docs/features/custom-field-serializer.md)\n- [Custom class (de)serializer](docs/features/custom-class-serializer.md)\n- [Flatten](docs/features/flatten.md)\n\n## LICENSE\n\nThis project is licensed under the [MIT license](https://github.com/yukinarit/pyserde/blob/master/LICENSE).\n',
    'author': 'yukinarit',
    'author_email': 'yukinarit84@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yukinarit/pyserde',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import dataclasses
import enum
import logging
import uuid
from typing import Dict, List, Optional, Tuple

import pytest

import serde

from . import data
from .common import (
    all_formats,
    format_dict,
    format_json,
    format_msgpack,
    format_toml,
    format_tuple,
    format_yaml,
    opt_case,
    opt_case_ids,
    type_ids,
    types,
)

log = logging.getLogger('test')

serde.init(True)


@pytest.mark.parametrize('t,T', types, ids=type_ids())
@pytest.mark.parametrize('opt', opt_case, ids=opt_case_ids())
@pytest.mark.parametrize('se,de', all_formats)
def test_simple(se, de, opt, t, T):
    log.info(f'Running test with se={se.__name__} de={de.__name__} opts={opt}')

    @serde.serde(**opt)
    class C:
        i: int
        t: T

    c = C(10, t)
    assert c == de(C, se(c))

    @serde.serde(**opt)
    class Nested:
        t: T

    @serde.serde(**opt)
    class C:
        n: Nested

    c = C(Nested(t))
    assert c == de(C, se(c))

    if se is not serde.toml.to_toml:
        assert t == de(T, se(t))


@pytest.mark.parametrize('t,T', types, ids=type_ids())
@pytest.mark.parametrize('opt', opt_case, ids=opt_case_ids())
@pytest.mark.parametrize('se,de', (format_dict + format_tuple))
def test_simple_with_reuse_instances(se, de, opt, t, T):
    log.info(f'Running test with se={se.__name__} de={de.__name__} opts={opt} while reusing instances')

    @serde.serde(**opt)
    class C:
        i: int
        t: T

    c = C(10, t)
    assert c == de(C, se(c, reuse_instances=True), reuse_instances=True)

    @serde.serde(**opt)
    class Nested:
        t: T

    @serde.serde(**opt)
    class C:
        n: Nested

    c = C(Nested(t))
    assert c == de(C, se(c, reuse_instances=True), reuse_instances=True)


def test_non_dataclass():
    @serde.serde
    class Foo:
        i: int

    @serde.serialize
    class Bar:
        i: int

    @serde.deserialize
    class Baz:
        i: int


# test_string_forward_reference_works currently only works with global visible classes
# and can not be mixed with PEP 563 "from __future__ import annotations"
@dataclasses.dataclass
class ForwardReferenceFoo:
    bar: 'ForwardReferenceBar'


@serde.serde
class ForwardReferenceBar:
    i: int


# assert type is str
assert 'ForwardReferenceBar' == dataclasses.fields(ForwardReferenceFoo)[0].type

# setup pyserde for Foo after Bar becomes visible to global scope
serde.serde(ForwardReferenceFoo)

# now the type really is of type Bar
assert ForwardReferenceBar == dataclasses.fields(ForwardReferenceFoo)[0].type
assert ForwardReferenceBar == next(serde.compat.dataclass_fields(ForwardReferenceFoo)).type


# verify usage works
def test_string_forward_reference_works():
    h = ForwardReferenceFoo(bar=ForwardReferenceBar(i=10))
    h_dict = {"bar": {"i": 10}}

    assert serde.to_dict(h) == h_dict
    assert serde.from_dict(ForwardReferenceFoo, h_dict) == h


# trying to use string forward reference normally will throw
def test_unresolved_forward_reference_throws():
    with pytest.raises(serde.SerdeError) as e:

        @serde.serde
        class UnresolvedForwardFoo:
            bar: 'UnresolvedForwardBar'

        @serde.serde
        class UnresolvedForwardBar:
            i: int

    assert "Failed to resolve type hints for UnresolvedForwardFoo" in str(e)


@pytest.mark.parametrize('opt', opt_case, ids=opt_case_ids())
@pytest.mark.parametrize('se,de', all_formats)
def test_list(se, de, opt):
    @serde.serde(**opt)
    class Variant:
        d: List

    # List can contain different types (except Toml).
    if se is not serde.toml.to_toml:
        p = Variant([10, 'foo', 10.0, True])
        assert p == de(Variant, se(p))


@pytest.mark.parametrize('opt', opt_case, ids=opt_case_ids())
@pytest.mark.parametrize('se,de', all_formats)
def test_dict(se, de, opt):
    from .data import PriDict

    if se in (serde.json.to_json, serde.msgpack.to_msgpack, serde.toml.to_toml):
        # JSON, Msgpack, Toml don't allow non string key.
        p = PriDict({'10': 10}, {'foo': 'bar'}, {'100.0': 100.0}, {'True': False})
        assert p == de(PriDict, se(p))
    else:
        p = PriDict({10: 10}, {'foo': 'bar'}, {100.0: 100.0}, {True: False})
        assert p == de(PriDict, se(p))

    @serde.serde(**opt)
    class Variant:
        d: Dict[int, str]

    p = Variant({'10': 10, 'foo': 'bar', '100.0': 100.0, 'True': False})
    assert p == de(Variant, se(p))


@pytest.mark.parametrize('opt', opt_case, ids=opt_case_ids())
@pytest.mark.parametrize('se,de', all_formats)
def test_enum(se, de, opt):
    from serde.compat import is_enum

    from .data import IE, E

    class Inner(enum.IntEnum):
        V0 = enum.auto()
        V1 = enum.auto()
        V2 = enum.auto()

    class NestedEnum(enum.Enum):
        V = Inner.V0

    @serde.serde(**opt)
    class Foo:
        a: E
        b: IE
        c: NestedEnum
        d: E = E.S
        e: IE = IE.V1
        f: NestedEnum = NestedEnum.V

    f = Foo(E.S, IE.V0, NestedEnum.V)
    ff = de(Foo, se(f))
    assert f == ff
    assert is_enum(ff.a) and isinstance(ff.a, E)
    assert is_enum(ff.b) and isinstance(ff.b, IE)
    assert is_enum(ff.c) and isinstance(ff.c, NestedEnum)
    assert is_enum(ff.d) and isinstance(ff.d, E)
    assert is_enum(ff.e) and isinstance(ff.e, IE)
    assert is_enum(ff.f) and isinstance(ff.f, NestedEnum)

    # pyserde automatically convert enum compatible value.
    f = Foo('foo', 2, Inner.V0, True, 10, Inner.V0)
    ff = de(Foo, se(f))
    assert is_enum(ff.a) and isinstance(ff.a, E) and ff.a == E.S
    assert is_enum(ff.b) and isinstance(ff.b, IE) and ff.b == IE.V1
    assert is_enum(ff.c) and isinstance(ff.c, NestedEnum) and ff.c == NestedEnum.V
    assert is_enum(ff.d) and isinstance(ff.d, E) and ff.d == E.B
    assert is_enum(ff.e) and isinstance(ff.e, IE) and ff.e == IE.V2
    assert is_enum(ff.f) and isinstance(ff.f, NestedEnum) and ff.f == NestedEnum.V


@pytest.mark.parametrize('se,de', all_formats)
def test_enum_imported(se, de):
    from .data import EnumInClass

    c = EnumInClass()
    cc = de(EnumInClass, se(c))
    assert c == cc


@pytest.mark.parametrize('opt', opt_case, ids=opt_case_ids())
@pytest.mark.parametrize('se,de', all_formats)
def test_tuple(se, de, opt):
    @serde.serde(**opt)
    class Homogeneous:
        i: Tuple[int, int]
        s: Tuple[str, str]
        f: Tuple[float, float]
        b: Tuple[bool, bool]

    p = Homogeneous((10, 20), ('a', 'b'), (10.0, 20.0), (True, False))
    assert p == de(Homogeneous, se(p))

    # List can also be used.
    p = Homogeneous([10, 20], ['a', 'b'], [10.0, 20.0], [True, False])
    assert p != de(Homogeneous, se(p))

    @serde.serde(**opt)
    class Variant:
        t: Tuple[int, str, float, bool]

    # Toml doesn't support variant type of array.
    if se is not serde.toml.to_toml:
        p = Variant((10, 'a', 10.0, True))
        assert p == de(Variant, se(p))

    @serde.serde(**opt)
    class BareTuple:
        t: Tuple

    p = BareTuple((10, 20))
    assert p == de(BareTuple, se(p))

    @serde.serde(**opt)
    class Nested:
        i: Tuple[data.Int, data.Int]
        s: Tuple[data.Str, data.Str]
        f: Tuple[data.Float, data.Float]
        b: Tuple[data.Bool, data.Bool]

    # hmmm.. Nested tuple doesn't work ..
    if se is not serde.toml.to_toml:
        p = Nested(
            (data.Int(10), data.Int(20)),
            (data.Str("a"), data.Str("b")),
            (data.Float(10.0), data.Float(20.0)),
            (data.Bool(True), data.Bool(False)),
        )
        assert p == de(Nested, se(p))


@pytest.mark.parametrize('se,de', all_formats)
def test_single_element_tuples(se, de):
    @serde.serde
    class Foo:
        a: Tuple[int]
        b: Tuple[uuid.UUID]

    foo = Foo(a=(1,), b=(uuid.UUID("855f07da-c3cd-46f2-9557-b8dbeb99ff42"),))
    assert serde.to_dict(foo) == {"a": foo.a, "b": foo.b}

    assert foo == de(Foo, se(foo))


@pytest.mark.parametrize('se,de', all_formats)
def test_dataclass_default_factory(se, de):
    @serde.serde
    class Foo:
        foo: str
        items: Dict[str, int] = serde.field(default_factory=dict)

    f = Foo('bar')
    assert f == de(Foo, se(f))

    assert {'foo': 'bar', 'items': {}} == serde.to_dict(f)
    assert f == serde.from_dict(Foo, {'foo': 'bar'})


@pytest.mark.parametrize('se,de', all_formats)
def test_default(se, de):
    from serde import from_dict, from_tuple

    from .data import OptDefault, PriDefault

    p = PriDefault()
    assert p == de(PriDefault, se(p))

    p = PriDefault()
    assert p == from_dict(PriDefault, {})
    assert p == from_dict(PriDefault, {'i': 10})
    assert p == from_dict(PriDefault, {'i': 10, 's': 'foo'})
    assert p == from_dict(PriDefault, {'i': 10, 's': 'foo', 'f': 100.0})
    assert p == from_dict(PriDefault, {'i': 10, 's': 'foo', 'f': 100.0, 'b': True})
    assert p == from_tuple(PriDefault, (10, 'foo', 100.0, True))

    o = OptDefault()
    assert o == de(OptDefault, se(o))

    o = OptDefault()
    assert o == from_dict(OptDefault, {})
    assert o == from_dict(OptDefault, {"n": None})
    assert o == from_dict(OptDefault, {"n": None, "i": 10})
    assert o == from_tuple(OptDefault, (None, 10))

    o = OptDefault(n=None, i=None)
    assert o == from_dict(OptDefault, {"n": None, "i": None})
    assert o == from_tuple(OptDefault, (None, None))

    assert 10 == dataclasses.fields(PriDefault)[0].default
    assert 'foo' == dataclasses.fields(PriDefault)[1].default
    assert 100.0 == dataclasses.fields(PriDefault)[2].default
    assert True is dataclasses.fields(PriDefault)[3].default


@pytest.mark.parametrize('se,de', (format_dict + format_tuple + format_json + format_msgpack + format_yaml))
def test_list_pri(se, de):
    p = [data.PRI, data.PRI]
    assert p == de(data.ListPri, se(p))

    p = []
    assert p == de(data.ListPri, se(p))


@pytest.mark.parametrize('se,de', (format_dict + format_tuple + format_json + format_msgpack + format_yaml))
def test_dict_pri(se, de):
    p = {'1': data.PRI, '2': data.PRI}
    assert p == de(data.DictPri, se(p))

    p = {}
    assert p == de(data.DictPri, se(p))


def test_json():
    p = data.Pri(10, 'foo', 100.0, True)
    s = '{"i": 10, "s": "foo", "f": 100.0, "b": true}'
    assert s == serde.json.to_json(p)

    assert '10' == serde.json.to_json(10)
    assert '[10, 20, 30]' == serde.json.to_json([10, 20, 30])
    assert '{"foo": 10, "fuga": 10}' == serde.json.to_json({'foo': 10, 'fuga': 10})


def test_msgpack():
    p = data.Pri(10, 'foo', 100.0, True)
    d = b'\x84\xa1i\n\xa1s\xa3foo\xa1f\xcb@Y\x00\x00\x00\x00\x00\x00\xa1b\xc3'
    assert d == serde.msgpack.to_msgpack(p)
    assert p == serde.msgpack.from_msgpack(data.Pri, d)


def test_msgpack_unnamed():
    p = data.Pri(10, 'foo', 100.0, True)
    d = b'\x94\n\xa3foo\xcb@Y\x00\x00\x00\x00\x00\x00\xc3'
    assert d == serde.msgpack.to_msgpack(p, named=False)
    assert p == serde.msgpack.from_msgpack(data.Pri, d, named=False)


@pytest.mark.parametrize('se,de', all_formats)
def test_rename(se, de):
    @serde.serde
    class Foo:
        class_name: str = serde.field(rename='class')

    f = Foo(class_name='foo')
    assert f == de(Foo, se(f))


@pytest.mark.parametrize('se,de', format_msgpack)
def test_rename_msgpack(se, de):
    @serde.serde(rename_all='camelcase')
    class Foo:
        class_name: str

    f = Foo(class_name='foo')
    assert f == de(Foo, se(f, named=True), named=True)
    assert f == de(Foo, se(f, named=False), named=False)


@pytest.mark.parametrize('se,de', (format_dict + format_json + format_yaml + format_toml))
def test_rename_formats(se, de):
    @serde.serde(rename_all='camelcase')
    class Foo:
        class_name: str

    f = Foo(class_name='foo')
    assert f == de(Foo, se(f))


@pytest.mark.parametrize('se,de', (format_dict + format_json + format_msgpack + format_yaml + format_toml))
def test_skip_if(se, de):
    @serde.serde
    class Foo:
        comments: Optional[List[str]] = serde.field(default_factory=list, skip_if=lambda v: len(v) == 0)
        attrs: Optional[Dict[str, str]] = serde.field(default_factory=dict, skip_if=lambda v: v is None or len(v) == 0)

    f = Foo(['foo'], {"bar": "baz"})
    assert f == de(Foo, se(f))

    f = Foo([])
    ff = de(Foo, se(f))
    assert ff.comments == []
    assert ff.attrs == {}


@pytest.mark.parametrize('se,de', all_formats)
def test_skip_if_false(se, de):
    @serde.serde
    class Foo:
        comments: Optional[List[str]] = serde.field(default_factory=list, skip_if_false=True)

    f = Foo(['foo'])
    assert f == de(Foo, se(f))


@pytest.mark.parametrize('se,de', (format_dict + format_json + format_msgpack + format_yaml + format_toml))
def test_skip_if_overrides_skip_if_false(se, de):
    @serde.serde
    class Foo:
        comments: Optional[List[str]] = serde.field(
            default_factory=list, skip_if_false=True, skip_if=lambda v: len(v) == 1
        )

    f = Foo(['foo'])
    ff = de(Foo, se(f))
    assert ff.comments == []


@pytest.mark.parametrize('se,de', all_formats)
def test_skip_if_default(se, de):
    @serde.serde
    class Foo:
        a: str = serde.field(default='foo', skip_if_default=True)

    f = Foo()
    assert f == de(Foo, se(f))

    assert serde.to_dict(Foo()) == {}
    assert serde.to_dict(Foo('bar')) == {'a': 'bar'}
    assert serde.from_dict(Foo, {}) == Foo()
    assert serde.from_dict(Foo, {'a': 'bar'}) == Foo('bar')


@pytest.mark.parametrize('se,de', format_msgpack)
def test_inheritance(se, de):
    @serde.serde
    class Base:
        a: int
        b: str

    @serde.serde
    class Derived(Base):
        c: float

    base = Base(10, "foo")
    assert base == de(Base, se(base))

    derived = Derived(10, "foo", 100.0)
    assert derived == de(Derived, se(derived))


@pytest.mark.parametrize('se,de', format_msgpack)
def test_ext(se, de):
    @serde.serde
    class Base:
        i: int
        s: str

    @serde.serde
    class DerivedA(Base):
        j: int

    @serde.serde
    class DerivedB(Base):
        k: float

    a = DerivedA(i=7, s="A", j=13)
    aa = de(Base, se(a))
    assert aa != a

    EXT_TYPE_DICT = {0: DerivedA, 1: DerivedB}
    # reverse the external type dict for faster serialization
    EXT_TYPE_DICT_REVERSED = {v: k for k, v in EXT_TYPE_DICT.items()}

    a = DerivedA(i=7, s="A", j=13)
    aa = de(None, se(a, ext_dict=EXT_TYPE_DICT_REVERSED), ext_dict=EXT_TYPE_DICT)
    assert aa == a

    b = DerivedB(i=3, s="B", k=11.0)
    bb = de(None, se(b, ext_dict=EXT_TYPE_DICT_REVERSED), ext_dict=EXT_TYPE_DICT)
    assert b == bb

    with pytest.raises(serde.SerdeError) as se_ex:
        se(a, ext_dict={})
    assert str(se_ex.value) == "Could not find type code for DerivedA in ext_dict"

    with pytest.raises(serde.SerdeError) as de_ex:
        de(None, se(a, ext_dict=EXT_TYPE_DICT_REVERSED), ext_dict={})
    assert str(de_ex.value) == "Could not find type for code 0 in ext_dict"


def test_exception_on_not_supported_types():
    class UnsupportedClass:
        def __init__(self):
            pass

    @serde.serde
    class Foo:
        b: UnsupportedClass

    with pytest.raises(serde.SerdeError) as se_ex:
        serde.to_dict(Foo(UnsupportedClass()))
    assert str(se_ex.value).startswith("Unsupported type: UnsupportedClass")

    with pytest.raises(serde.SerdeError) as de_ex:
        serde.from_dict(Foo, {"b": UnsupportedClass()})
    assert str(de_ex.value).startswith("Unsupported type: UnsupportedClass")


def test_dataclass_inheritance():
    @serde.serde
    class Base:
        i: int
        s: str

    @serde.serde
    class DerivedA(Base):
        j: int

    @serde.serde
    class DerivedB(Base):
        k: float

    # each class should have own scope
    # ensure the generated code of DerivedB does not overwrite the earlier generated code from DerivedA
    assert getattr(Base, serde.core.SERDE_SCOPE) is not getattr(DerivedA, serde.core.SERDE_SCOPE)
    assert getattr(DerivedA, serde.core.SERDE_SCOPE) is not getattr(DerivedB, serde.core.SERDE_SCOPE)

    base = Base(i=0, s="foo")
    assert base == serde.from_dict(Base, serde.to_dict(base))

    a = DerivedA(i=0, s="foo", j=42)
    assert a == serde.from_dict(DerivedA, serde.to_dict(a))

    b = DerivedB(i=0, s="foo", k=42.0)
    assert b == serde.from_dict(DerivedB, serde.to_dict(b))


def make_serde(class_name: str, se: bool, fields, *args, **kwargs):
    if se:
        return serde.de.deserialize(serde.se._make_serialize(class_name, fields, *args, **kwargs), **kwargs)
    else:
        return serde.se.serialize(serde.de._make_deserialize(class_name, fields, *args, **kwargs), **kwargs)


@pytest.mark.parametrize('se', (True, False))
def test_make_serialize_deserialize(se):
    fields = [('i', int, dataclasses.field())]
    Foo = make_serde('Foo', se, fields)

    f = Foo(10)
    assert serde.to_dict(f) == {'i': 10}
    assert str(f) == 'Foo(i=10)'
    assert serde.from_dict(Foo, {'i': 10}) == f

    # Test class attribute
    fields = [('int_field', int, dataclasses.field())]
    Foo = make_serde('Foo', se, fields, rename_all='pascalcase')
    f = Foo(10)
    assert serde.to_dict(f) == {'IntField': 10}

    # Test field attribute
    fields = [('i', int, dataclasses.field(metadata={'serde_skip': True})), ('j', float, dataclasses.field())]
    Foo = make_serde('Foo', se, fields)
    f = Foo(10, 100.0)
    assert serde.to_dict(f) == {'j': 100.0}

    # Test class/field attributes at the same time
    fields = [
        ('int_field', int, dataclasses.field(metadata={'serde_rename': 'renamed_field'})),
        ('float_field', float, dataclasses.field()),
    ]
    Foo = make_serde('Foo', se, fields, rename_all='pascalcase')
    f = Foo(10, 100.0)
    assert serde.to_dict(f) == {'renamed_field': 10, 'FloatField': 100.0}

    # Nested
    fields = [('v', int, dataclasses.field())]
    Bar = make_serde('Bar', se, fields)

    fields = [('bar', Bar, dataclasses.field())]
    Foo = make_serde('Foo', se, fields)
    f = Foo(Bar(10))
    assert serde.to_dict(f) == {'bar': {'v': 10}}
    assert serde.from_dict(Foo, {'bar': {'v': 10}}) == f


def test_exception_to_from_obj():
    @serde.serde
    class Foo:
        a: int

    class Bar:
        pass

    with pytest.raises(serde.SerdeError):
        serde.from_dict(Foo, {})



from decimal import Decimal

from serde import serde
from serde.json import from_json, to_json


@serde
class Foo:
    v: Decimal


def main():
    foo = Foo(Decimal(0.1))
    print(f"Into Json: {to_json(foo)}")

    s = '{"v": "0.1000000000000000055511151231257827021181583404541015625"}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



from typing import Optional
from serde import serde, from_dict, field

@serde
class Foo:
    bar: Optional[list[int]] = field(default_factory=list)

if __name__ == "__main__":
    print(from_dict(Foo, {"bar": None}))



from serde import serde, to_dict, from_dict, to_tuple, from_tuple
from typing import Optional, Union


@serde
class OptDefault:
    """
    Optionals.
    """

    n: Optional[int] = None
    i: Optional[int] = 10


@serde
class Foo:
    a: Union[int, float]


if __name__ == "__main__":
    """
    o = OptDefault()
    data = to_dict(o)
    print(data)
    o = from_dict(OptDefault, data)
    print(o)
    data = to_tuple(o)
    print(data)
    o = from_tuple(OptDefault, data)
    print(o)
    """

    f = from_dict(Foo, {"a": "10"})
    print(f)
